import os
import subprocess
import time
from contextlib import contextmanager
from datetime import timedelta
from unittest.mock import patch

from django.test import TestCase

from parviraptor.worker import QueueWorker, QueueWorkerLogger

from .models import DummyJob
from .utils import disable_logging


class QueueTestCase(TestCase):
    """Tests für `QueueWorker`.

    Ein Problem beim Testen des Workers ist, dass `QueueWorker.run()` im
    Wesentlichen eine Endlosschleife ist. Dazu patch-en wir `threading.Event`
    mit `ExplodingEvent`, damit nur eine definierte Anzahl an `wait`-Aufrufen
    möglich ist und danach der QueueWorker durch eine
    `MaxCallsReached`-Exception beendet wird.

    Standardmäßig (in `setUp`) werden die Pausen für "leere Queue" (1s) und
    "temporärer Fehler" (100s) stark voneinander abweichend konfiguriert,
    v.a. unter Berücksichtigung der maximalen Wait-Zahl (9). Das ermöglicht mit
    `assert_waits` exakt zu prüfen, wie oft welcher Fall aufgetreten ist.
    """

    def setUp(self):
        self.pause_if_queue_empty = timedelta(seconds=1)
        self.max_wait_calls = 9

    def test_sleep_if_queue_empty(self):
        self.run_worker()
        self.assert_waits(10, 0)

    def test_sleep_if_queue_only_contains_squashed_and_processed(self):
        DummyJob.objects.create(a=0, b=0, status=DummyJob.Status.PROCESSED)
        DummyJob.objects.create(a=0, b=0, status=DummyJob.Status.SQUASHED)
        DummyJob.objects.create(a=0, b=0, status=DummyJob.Status.SQUASHED)
        DummyJob.objects.create(a=0, b=0, status=DummyJob.Status.PROCESSED)
        self.run_worker()
        self.assert_waits(10, 0)

    @disable_logging()
    def test_successful_processing(self):
        job_a = DummyJob.objects.create(a=1, b=2)
        job_b = DummyJob.objects.create(a=3, b=7)

        self.run_worker()
        modification_date_before = job_a.modification_date  # ein Job reicht
        job_a.refresh_from_db()
        job_b.refresh_from_db()
        modification_date_after = job_a.modification_date
        self.assertGreater(modification_date_after, modification_date_before)

        self.assertEquals(job_a.status, DummyJob.Status.PROCESSED)
        self.assertEquals(job_a.result, 3)
        self.assertEquals(job_b.status, DummyJob.Status.PROCESSED)
        self.assertEquals(job_b.result, 10)

    @disable_logging()
    def test_retry_on_temporary_failure(self):
        DummyJob.objects.create(a=0, b=1)  # schlägt fünf Mal fehl
        self.run_worker()
        job = DummyJob.objects.get()
        self.assertEqual(5, job.error_count)
        self.assertEqual(1, job.result)
        self.assertIsNone(job.error_message)
        self.assertEqual(DummyJob.Status.PROCESSED, job.status)
        self.assert_waits(5, 5)

    @disable_logging()
    def test_exceed_temporary_failure_threshold(self):
        DummyJob.objects.create(a=0, b=100000)
        self.run_worker()  # always fails temporarily
        job = DummyJob.objects.get()
        self.assertEqual(10, job.error_count)
        self.assertIsNone(job.result)
        self.assertEqual(DummyJob.Status.NEW, job.status)

    @disable_logging()
    def test_retry_on_not_processable(self):
        DummyJob.objects.create(a=0, b=1)
        with patch.object(DummyJob, "is_processable", lambda self: False):
            self.run_worker()
        job = DummyJob.objects.get()
        self.assertEquals(DummyJob.Status.NEW, job.status)
        self.assertEquals(0, job.error_count)
        self.assertIsNone(job.result)
        self.assert_waits(10, 0)

    @disable_logging()
    def test_retry_on_temporary_failure_calculates_backoff_properly(self):
        self.pause_if_queue_empty = timedelta(seconds=0)

        with patch("tests.models.MAX_ERROR_COUNT", 10):
            DummyJob.objects.create(a=0, b=1)
            self.run_worker()

        job = DummyJob.objects.get()
        self.assertEquals(10, job.error_count)
        self.assertEquals(1, job.result)
        self.assert_waits(0, 10)

        # Zur Nachvollziehbarkeit:
        temporary_failure_latency = 60 * (
            # die ersten 6 Fehlschläge geht es exponentiell hoch
            2**0
            + 2**1
            + 2**2
            + 2**3
            + 2**4
            + 2**5
            # ab jetzt aber nur noch konstant
            + 2**5
            + 2**5
            + 2**5
            + 2**5
        )

        self.assertEquals(
            temporary_failure_latency,
            self.exploding_event.total_wait_timeouts,
        )

    @disable_logging()
    def test_status_failed_on_exception(self):
        job = DummyJob.objects.create(a=1, b=0)  # wirft `ValueError`
        self.run_worker()
        job.refresh_from_db()
        self.assertEquals(job.status, DummyJob.Status.FAILED)

    @disable_logging()
    def test_status_failed_on_invalidjoberror(self):
        job = DummyJob.objects.create(a=50, b=50)  # wirft `InvalidJobError`
        self.run_worker()
        job.refresh_from_db()
        self.assertEqual(DummyJob.Status.FAILED, job.status)
        self.assertEqual("Ignoring result 100", job.error_message)

    @disable_logging()
    def test_status_ignored_on_ignorejob(self):
        job = DummyJob.objects.create(a=100, b=100)  # wirft `IgnoreJob`
        self.run_worker()
        job.refresh_from_db()
        self.assertEqual(DummyJob.Status.IGNORED, job.status)
        self.assertEqual("Ignoring result 200", job.error_message)

    @disable_logging()
    def test_status_deferred_on_deferjob(self):
        job = DummyJob.objects.create(a=150, b=150)  # wirft `DeferJob`
        self.run_worker()
        job.refresh_from_db()
        self.assertEqual(DummyJob.Status.DEFERRED, job.status)
        self.assertEqual("Deferring result 300", job.error_message)

    @disable_logging()
    def test_job_changes_get_saved_on_success_and_failure(self):
        # Wir können das Speichern bei temporären Fehlern nicht testen, da wir
        # die Queue-Verarbeitung abbrechen müssen, bevor der Job erfolgreich
        # verarbeitet wird. Siehe
        # `test_job_changes_get_saved_on_temporary_failure`.
        job_a = DummyJob.objects.create(a=1, b=2)  # keine Fehler
        job_b = DummyJob.objects.create(a=2, b=0)  # Fehler
        self.run_worker()
        for job in [job_a, job_b]:
            job.refresh_from_db()
            self.assertNotEquals(DummyJob.Status.NEW, job.status)
            self.assertNotEquals(None, job.result)

    @disable_logging()
    def test_fifo_depending_jobs_are_set_to_failed(self):
        job_a = DummyJob.objects.create(a=1, b=2)  # keine Fehler
        job_b = DummyJob.objects.create(a=1, b=0)  # Fehler
        job_c = DummyJob.objects.create(a=1, b=2)  # c und d hängen von job_b ab
        job_d = DummyJob.objects.create(a=1, b=2)  # → werden dann auch FAILED

        for job in [job_a, job_b, job_c, job_d]:
            job.refresh_from_db()
            self.assertEquals(DummyJob.Status.NEW, job.status)
        self.run_worker()
        for job in [job_a, job_b, job_c, job_d]:
            job.refresh_from_db()
        self.assertEqual(DummyJob.Status.PROCESSED, job_a.status)
        self.assertEqual(DummyJob.Status.FAILED, job_b.status)
        self.assertEqual("b cannot be 0", job_b.error_message)
        self.assertEqual(DummyJob.Status.FAILED, job_c.status)
        self.assertEqual("dependent jobs failed", job_c.error_message)
        self.assertEqual(DummyJob.Status.FAILED, job_d.status)
        self.assertEqual("dependent jobs failed", job_d.error_message)

    @disable_logging()
    def test_job_changes_get_saved_on_temporary_failure(self):
        self.max_wait_calls = 0
        job = DummyJob.objects.create(a=0, b=2)  # temporärer Fehler
        self.run_worker()
        job.refresh_from_db()
        self.assertEquals(DummyJob.Status.NEW, job.status)
        self.assertEquals(2, job.result)

    @disable_logging()
    def test_sigterm_handling(self):
        job_a = DummyJob.objects.create(a=-1, b=-1)  # sendet SIGTERM
        job_b = DummyJob.objects.create(a=1, b=2)
        self.run_worker()
        self.assert_waits(0, 0)

        # Der erste Job wurde vollständig verarbeitet, da das Signal-Handling
        # immer nur zwischen Jobs greift.
        job_a.refresh_from_db()
        self.assertEquals(DummyJob.Status.PROCESSED, job_a.status)

        # Der zweite Job hat immer noch den Status NEW:
        job_b.refresh_from_db()
        self.assertEquals(DummyJob.Status.NEW, job_b.status)
        self.assertEquals(None, job_b.result)

    def test_logging(self):
        logger = QueueWorkerLogger()
        with self.assertLogs() as cm:
            # keine Jobs offen
            logger.mutate_to_idle_state()
            # verarbeitet 5 Jobs
            for _ in range(5):
                logger.mutate_to_processing_state()
            # Für längere Zeit keine Jobs mehr offen ergibt nur 1 Meldung.
            # Dadurch dass wir explizit loggen, wenn wir wieder Jobs
            # verarbeiten, müssen wir "es ist nichts zu tun" nicht ständig
            # loggen in längeren Leerlaufphasen.
            logger.mutate_to_idle_state()
            logger.mutate_to_idle_state()
            logger.mutate_to_idle_state()
            logger.mutate_to_idle_state()
            logger.mutate_to_idle_state()
            # verarbeitet nochmal 4 Jobs
            for _ in range(4):
                logger.mutate_to_processing_state()
            # keine Jobs mehr offen
            logger.mutate_to_idle_state()
            logger.mutate_to_idle_state()
            # darf keine Jobs mehr verarbeiten
            logger.mutate_to_unprocessable_state()
            logger.mutate_to_unprocessable_state()
            logger.mutate_to_unprocessable_state()
            logger.mutate_to_unprocessable_state()
            # verarbeitet einen Job
            logger.mutate_to_processing_state()
            # darf keine Jobs mehr verarbeiten
            logger.mutate_to_unprocessable_state()
            logger.mutate_to_unprocessable_state()
            logger.mutate_to_unprocessable_state()
            logger.mutate_to_unprocessable_state()
            # keine Jobs mehr offen
            logger.mutate_to_idle_state()
            logger.mutate_to_idle_state()
        self.assertEqual(
            [
                "processed 0 jobs. awaiting new jobs.",
                "processing jobs...",
                "processed 5 jobs. awaiting new jobs.",
                "processing jobs...",
                "processed 4 jobs. awaiting new jobs.",
                "processed 0 jobs so far.",
                "queue turned unprocessable right now. "
                + "waiting until queue is processable again.",
                "processing jobs...",
                "processed 1 jobs so far.",
                "queue turned unprocessable right now. "
                + "waiting until queue is processable again.",
                "processed 0 jobs. awaiting new jobs.",
            ],
            [record.message for record in cm.records],
        )

    @disable_logging()
    def test_sigterm_interrupts_sleep(self):
        with self.assert_max_runtime(timedelta(seconds=2)):
            self.delayed_send_sigterm(os.getpid(), timedelta(seconds=1))
            worker = QueueWorker(
                DummyJob, pause_if_queue_empty=timedelta(seconds=3)
            )
            worker.run()

    @disable_logging()
    def test_sigint_interrupts_sleep(self):
        with self.assert_max_runtime(timedelta(seconds=2)):
            self.delayed_send_sigint(os.getpid(), timedelta(seconds=1))
            worker = QueueWorker(
                DummyJob, pause_if_queue_empty=timedelta(seconds=3)
            )
            worker.run()

    @contextmanager
    def assert_max_runtime(self, max_runtime):
        t0 = time.time()
        yield
        t1 = time.time()
        self.assertLessEqual(t1 - t0, max_runtime.seconds)

    def delayed_send_sigterm(self, pid, delay):
        subprocess.Popen(["sh", "-c", f"sleep {delay.seconds} && kill {pid}"])

    def delayed_send_sigint(self, pid, delay):
        subprocess.Popen(
            ["sh", "-c", f"sleep {delay.seconds} && kill -2 {pid}"]
        )

    def run_worker(self):
        with self.exploding_event():
            self.worker = QueueWorker(
                DummyJob,
                pause_if_queue_empty=self.pause_if_queue_empty,
            )
            try:
                self.worker.run()
            except MaxCallsReached:
                pass

    @contextmanager
    def exploding_event(self):
        self.exploding_event = ExplodingEvent(self.max_wait_calls)
        with patch("threading.Event", lambda: self.exploding_event):
            yield

    def assert_waits(self, empty_queue_count, temporary_failure_count):
        temporary_failure_latency = 60 * sum(
            [2 ** min(x, 5) for x in range(temporary_failure_count)]
        )

        expected = (
            empty_queue_count * self.pause_if_queue_empty.seconds
            + temporary_failure_latency
        )
        self.assertEquals(expected, self.exploding_event.total_wait_timeouts)


class ExplodingEvent:
    """Mock-Objekt für `threading.Event`.

    Speichert alle Aufrufe von `wait` und wirft `MaxCallsReached`, sobald eine
    festgelegte Zahl an Aufrufen erfolgt ist.
    """

    def __init__(self, max_waits):
        self.flag = False
        self.max_waits = max_waits
        self.wait_timeouts = []

    def is_set(self):
        return self.flag

    def set(self):
        self.flag = True

    def clear(self):
        self.flag = False

    def wait(self, timeout=None):
        self.wait_timeouts.append(timeout if timeout is not None else 0)
        if len(self.wait_timeouts) > self.max_waits:
            raise MaxCallsReached("wait()", self.wait_timeouts)

    @property
    def total_wait_timeouts(self):
        return sum(self.wait_timeouts)


class MaxCallsReached(Exception):
    def __init__(self, name, calls):
        super().__init__(f"{name} was called too often: {len(calls)} times")



from datetime import datetime, timedelta
from io import StringIO
from unittest.mock import patch

from django.core.management import call_command
from django.test import TransactionTestCase
from django.utils import timezone

from tests.models import DummyJob, DummyProductJob, IncrementCounterJob

COMMAND_MODULE = "parviraptor.management.commands.clean_old_finished_jobs"


class CleanOldFinishedJobsTests(TransactionTestCase):
    def test_can_process_all_queues(self):
        jobs = []

        def patched(Job, max_age_in_days):
            self.assertEqual(1, max_age_in_days)
            jobs.append(Job)

        with patch(
            COMMAND_MODULE + "._delete_old_finished_jobs",
            patched,
        ):
            self._call_command("1", "--all-queues")

        self.assertCountEqual(
            [DummyJob, DummyProductJob, IncrementCounterJob], jobs
        )

    def test_can_process_single_queue(self):
        jobs = []

        def patched(Job, max_age_in_days):
            self.assertEqual(1, max_age_in_days)
            jobs.append(Job)

        with patch(
            COMMAND_MODULE + "._delete_old_finished_jobs",
            patched,
        ):
            self._call_command("1", "--queue=tests.DummyJob")

        self.assertEqual([DummyJob], jobs)

    def test_leaves_unfinished_old_jobs_untouched(self):
        old_date = datetime.now(tz=timezone.utc) - timedelta(days=23)

        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="FAILED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.all().update(modification_date=old_date)
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")

        self.assertEqual(5, DummyJob.objects.count())
        self._call_command("1", "--queue=tests.DummyJob")
        self.assertEqual(4, DummyJob.objects.count())

    def test_no_jobs_to_be_deleted(self):
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="FAILED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")

        self.assertEqual(5, DummyJob.objects.count())
        self._call_command("1", "--queue=tests.DummyJob")
        self.assertEqual(5, DummyJob.objects.count())

    def test_can_delete_all_old_finished_jobs(self):
        old_date = datetime.now(tz=timezone.utc) - timedelta(days=23)

        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.all().update(modification_date=old_date)
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")

        self.assertEqual(5, DummyJob.objects.count())
        self._call_command("1", "--queue=tests.DummyJob")
        self.assertEqual(2, DummyJob.objects.count())

    def _call_command(self, *args):
        DEV_NULL = StringIO()
        call_command("clean_old_finished_jobs", *args, stdout=DEV_NULL)



from datetime import datetime

from django.test import TestCase
from django.utils import timezone

from .models import DummyJob, DummyProductJob


class AbstractJobTests(TestCase):
    """
    Die Tests hier beziehen sich auf Funktionalität des `AbstractJob`.
    Da der `AbstractJob` bekanntlich ein abstraktes Model ist,
    nutzen wir u.a. den `DummyJob` zum Testen.
    """

    def test_count_failed_jobs(self):
        self.assertEqual(0, DummyJob.count_failed_jobs())
        job = DummyJob.objects.create(a=1, b=1)
        self.assertEqual(0, DummyJob.count_failed_jobs())
        job.status = DummyJob.Status.FAILED
        job.save()
        self.assertEqual(1, DummyJob.count_failed_jobs())

    def test_count_long_processing_jobs(self):
        self.assertEqual(0, DummyJob.count_long_processing_jobs())

        job = DummyJob.objects.create(a=1, b=1)
        self.assertEqual(0, DummyJob.count_long_processing_jobs())

        DummyJob.objects.filter(pk=job.id).update(
            status=DummyJob.Status.PROCESSING
        )
        self.assertEqual(0, DummyJob.count_long_processing_jobs())

        DummyJob.objects.filter(pk=job.id).update(
            modification_date=datetime(
                2000,
                1,
                1,
                13,
                37,
                42,
                tzinfo=timezone.utc,
            )
        )
        self.assertEqual(1, DummyJob.count_long_processing_jobs())

    def test_can_tell_disjoint_queues(self):
        DummyProductJob.objects.bulk_create(
            [
                DummyProductJob(
                    shop_name=shop_name,
                    product_name=product_name,
                    action=action,
                )
                for shop_name in ["shop-a", "shop-b"]
                for product_name in ["prod-a", "prod-b"]
                for action in ["foo", "bar"]
            ]
        )
        self.assertCountEqual(
            [
                {"shop_name": "shop-a", "product_name": "prod-a"},
                {"shop_name": "shop-a", "product_name": "prod-b"},
                {"shop_name": "shop-b", "product_name": "prod-a"},
                {"shop_name": "shop-b", "product_name": "prod-b"},
            ],
            DummyProductJob.get_queryset_filters_for_disjoint_queues(),
        )



import logging
from datetime import datetime, timedelta

from django.apps import apps
from django.core.management.base import BaseCommand
from django.utils import timezone

from parviraptor.models.abstract import AbstractJob
from parviraptor.utils import enumerate_job_models, iter_chunks

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Cleans up obsolete finished jobs"

    def add_arguments(self, parser):
        parser.add_argument(
            "max_age_in_days",
            type=int,
            help=(
                """
                Jobs which are as old or older than the given age in
                days are removed. If there is a not successfully finished
                job within the resulting timeframe, only the jobs until the
                first failed job are removed if there are some.
                """
            ),
        )
        queue_switch = parser.add_mutually_exclusive_group()
        queue_switch.add_argument(
            "--all-queues",
            action="store_true",
            help=(
                """
                Removes old finished jobs in all job models existing
                within the current installation.
                """
            ),
        )
        queue_switch.add_argument(
            "--queue",
            type=str,
            help="<app_label>.<ModelName>, e.g. my_app.SomeRandomJob",
        )

    def handle(self, max_age_in_days: int, **options):
        relevant_job_models = (
            enumerate_job_models()
            if options["all_queues"]
            else [_load_model_from_fully_qualified_name(options["queue"])]
        )
        for Job in relevant_job_models:
            _delete_old_finished_jobs(Job, max_age_in_days)


def _delete_old_finished_jobs(Job: type[AbstractJob], max_age_in_days: int):
    border = datetime.now(tz=timezone.utc) - timedelta(days=max_age_in_days)
    old_jobs = Job.objects.filter(modification_date__lte=border)
    if maybe_unfinished_job := old_jobs.exclude(
        status__in=("PROCESSED", "DEFERRED", "IGNORED")
    ).first():
        _delete_in_chunks(Job, old_jobs.filter(id__lt=maybe_unfinished_job.pk))
    else:
        _delete_in_chunks(Job, old_jobs)


def _delete_in_chunks(Job: type[AbstractJob], jobs):
    pks = jobs.values_list("pk", flat=True)
    pks_count = pks.count()

    CHUNK_SIZE = 2_000
    full_chunks = pks_count // CHUNK_SIZE
    remainder = 0 if pks_count % CHUNK_SIZE == 0 else 1
    chunk_count = full_chunks + remainder

    for i, chunk in enumerate(iter_chunks(CHUNK_SIZE, pks), start=1):
        logger.info(f"{Job.__name__}: processing chunk {i}/{chunk_count}")
        Job.objects.filter(pk__in=chunk).delete()


def _load_model_from_fully_qualified_name(name: str) -> type[AbstractJob]:
    app_label, model_name = name.split(".")
    return apps.get_model(app_label, model_name)



from django.shortcuts import render
from django.views.generic import View

from ..models.abstract import JobStatus
from ..utils import enumerate_job_models


class OpenQueueEntriesView(View):
    def get(self, request):
        context = self.get_context()
        return render(request, "parviraptor/open_queue_entries.html", context)

    def get_context(self) -> dict:
        open_queue_entries = {
            job_class.__name__: job_class.objects.filter(
                status__in=[JobStatus.NEW, JobStatus.PROCESSING],
            ).count()
            for job_class in enumerate_job_models()
        }

        return {"open_queue_entries": open_queue_entries}



import logging
from contextlib import contextmanager
from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from parviraptor.worker import QueueWorker

logger = logging.getLogger(__name__)


@contextmanager
def queue_worker_mock(log):
    run = QueueWorker.run
    QueueWorker.run = lambda *x, **y: logger.debug(log)
    yield
    QueueWorker.run = run


class ProcessQueueTestCase(TestCase):
    def test_process_queue(self):
        log = "Are you mocking me?"
        # QueueWorker.run wird in test_queue getestet
        with queue_worker_mock(log):
            # nicht existierendes Model
            out = StringIO()
            self.assertRaises(
                LookupError,
                lambda: call_command(
                    "process_queue",
                    ["tests", "NoneExistingModel"],
                    stdout=out,
                ),
            )
            self.assertNotIn("QueueWorker for", out.getvalue())

            # existierendes Model
            out = StringIO()
            with self.assertLogs(logger="", level="DEBUG") as cm:
                call_command("process_queue", ["tests", "DummyJob"], stdout=out)
            self.assertIn("QueueWorker for DummyJob", out.getvalue())
            # prüfen, dass QueueWorker.run auch wirklich aufgerufen wird
            self.assertIn(log, cm.output[0])



from xml.etree import ElementTree as ET

from django.test import TestCase

from .models import DummyJob, IncrementCounterJob


class OpenQueueEntriesTests(TestCase):
    def parse_records(self):
        resp = self.client.get("/open-queue-entries/")
        self.assertEqual(200, resp.status_code)

        content = resp.content.decode()
        table = ET.fromstring(content).find("body/table")
        rows = table.findall("tr")

        # die erste Tabellenzeile sind die Überschriften
        self.assertEqual({"th"}, {el.tag for el in rows[0]})

        return {(row[0].text, row[1].text) for row in rows[1:]}

    def test_empty_queues(self):
        self.assertEqual(
            {
                ("DummyJob", "0"),
                ("DummyProductJob", "0"),
                ("IncrementCounterJob", "0"),
            },
            self.parse_records(),
        )

    def test_shows_open_and_processing_queue_entries(self):
        DummyJob.objects.create(a=1, b=2)
        self.assertEqual(
            {
                ("DummyJob", "1"),
                ("DummyProductJob", "0"),
                ("IncrementCounterJob", "0"),
            },
            self.parse_records(),
        )
        DummyJob.objects.create(a=1, b=2)
        self.assertEqual(
            {
                ("DummyJob", "2"),
                ("DummyProductJob", "0"),
                ("IncrementCounterJob", "0"),
            },
            self.parse_records(),
        )
        DummyJob.objects.create(a=1, b=2, status="PROCESSING")
        self.assertEqual(
            {
                ("DummyJob", "3"),
                ("DummyProductJob", "0"),
                ("IncrementCounterJob", "0"),
            },
            self.parse_records(),
        )
        DummyJob.objects.create(a=1, b=2, status="PROCESSED")
        self.assertEqual(
            {
                ("DummyJob", "3"),
                ("DummyProductJob", "0"),
                ("IncrementCounterJob", "0"),
            },
            self.parse_records(),
        )
        IncrementCounterJob.objects.create(counter_id="irrelevant")
        self.assertEqual(
            {
                ("DummyJob", "3"),
                ("DummyProductJob", "0"),
                ("IncrementCounterJob", "1"),
            },
            self.parse_records(),
        )



"""A simple test framework for parviraptor-based applications."""

from .case import QueueTestCase
from .factory import make_test_case_for_all_queues

__all__ = [
    "QueueTestCase",
    "make_test_case_for_all_queues",
]



import re
from unittest.mock import patch

from django.test import TestCase


class MonitoringEndpointTests(TestCase):
    def request_monitoring_endpoint(self):
        resp = self.client.get("/queue-monitoring/")
        content = resp.content.decode()
        normalized = re.sub(r"\s+", " ", content).strip()
        return normalized

    def test_does_not_complain_if_queue_is_empty(self):
        self.assertEqual("<p>Alles OK</p>", self.request_monitoring_endpoint())

    def test_complains_about_failed_jobs(self):
        with patch(
            "tests.models.DummyJob.count_failed_jobs",
            lambda: 1,
        ):
            self.assertEqual(
                "<p> There are 1 failed, 0 long processing and 0 long "
                "unprocessed DummyJobs. </p>",
                self.request_monitoring_endpoint(),
            )

    def test_complains_about_long_processing_jobs(self):
        with patch(
            "tests.models.DummyJob.count_long_processing_jobs",
            lambda: 1,
        ):
            self.assertEqual(
                "<p> There are 0 failed, 1 long processing and 0 long "
                "unprocessed DummyJobs. </p>",
                self.request_monitoring_endpoint(),
            )

    def test_complains_about_long_unprocessed_jobs(self):
        with patch(
            "tests.models.DummyJob.count_long_unprocessed_jobs",
            lambda: 1,
        ):
            self.assertEqual(
                "<p> There are 0 failed, 0 long processing and 1 long "
                "unprocessed DummyJobs. </p>",
                self.request_monitoring_endpoint(),
            )



from django.apps import apps
from django.core.management.base import BaseCommand

from parviraptor.worker import QueueWorker


class Command(BaseCommand):
    help = "Call QueueWorker for a specific model"

    def add_arguments(self, parser):
        parser.add_argument(
            "app_label",
            type=str,
        )
        parser.add_argument(
            "model_name",
            type=str,
        )

    def handle(self, app_label, model_name, *args, **options):
        Model = apps.get_model(app_label, model_name)
        worker = QueueWorker(Model)

        self.stdout.write(f"Starting QueueWorker for {model_name} ...")

        worker.run()

        self.stdout.write(f"Stopping QueueWorker for {model_name}")



from ..utils import enumerate_job_models
from .case import QueueTestCase


def make_test_case_for_all_queues(**static_fields) -> type[QueueTestCase]:
    """
    Infers a test case for all non-abstract parviraptor job models within
    the current Django environment.

    It covers processing jobs both sequentially and in parallel, and ensures
    job instances exist for each concrete job class. Therefore, you should
    derive from the inferred base class and extend `setUp()` to create job
    instances.

    Parameters:
    - `static_fields`: keyword parameters. they are directly passed as static
      members to the test class. One might want to pass `maxDiff=None` and
      `fixtures=["some-fixture.json"]`, for example.
    """
    model_classes = enumerate_job_models()

    class _TestCase(QueueTestCase):
        queues = [model_class.__name__ for model_class in model_classes]

        def test_can_process_queue_sequentially(self):
            self._process_all_queues(num_workers=1)

        def test_can_process_queue_concurrently(self):
            self._process_all_queues(num_workers=8)

        def _process_all_queues(self, num_workers: int):
            for model_class in model_classes:
                with self.subTest(model_class.__name__):
                    self.assertGreater(model_class.objects.count(), 0)
                    self.process_queue(
                        model_class,
                        model_class.objects.all(),
                        num_workers,
                        create_jobs=False,
                    )
                    self._assert_jobs_are_processed_in_proper_order(model_class)

        def _assert_jobs_are_processed_in_proper_order(self, model_class):
            for f in model_class.get_queryset_filters_for_disjoint_queues():
                jobs = model_class.objects.filter(**f)
                ordered_ids = self.get_ordered_ids(jobs, "pk")
                ids_in_order_of_processing = self.get_ordered_ids(
                    jobs, "modification_date"
                )
                self.assertEqual(ordered_ids, ids_in_order_of_processing)

    for field, value in static_fields.items():
        setattr(_TestCase, field, value)
    return _TestCase



from dataclasses import dataclass
from typing import Iterable, List

from parviraptor.models.abstract import AbstractJob


@dataclass(frozen=True)
class QueueMonitoringResult:
    queue_name: str
    failed_jobs_count: int
    long_processing_jobs_count: int
    long_unprocessed_jobs_count: int


def monitor_queue_entries(
    job_classes: Iterable[type],
) -> List[QueueMonitoringResult]:
    """
    Returns `QueueMonitoringResult` per passed job class.

    A `QueueMonitoringResult` is returned per job class if at least one of the
    following conditions is met:
    - jobs failed (failed_jobs_count > 0)
    - backlog is potentially too large (long_unprocessed_jobs_count > 0)
    - jobs potentially crashed (long_processing_jobs_count > 1)
    """

    mistyped_classes = list(
        filter(
            lambda job_class: not issubclass(job_class, AbstractJob),
            job_classes,
        )
    )

    if mistyped_classes:
        raise TypeError(mistyped_classes)

    results = []

    for job_class in job_classes:
        failed_jobs_count = job_class.count_failed_jobs()
        long_processing_jobs_count = job_class.count_long_processing_jobs()
        long_unprocessed_jobs_count = job_class.count_long_unprocessed_jobs()

        if (
            failed_jobs_count
            or long_processing_jobs_count
            or long_unprocessed_jobs_count
        ):
            result = QueueMonitoringResult(
                queue_name=job_class.__name__,
                failed_jobs_count=failed_jobs_count,
                long_processing_jobs_count=long_processing_jobs_count,
                long_unprocessed_jobs_count=long_unprocessed_jobs_count,
            )
            results.append(result)
    return results



from unittest import TestCase

from parviraptor.test import make_test_case_for_all_queues

from .models import Counter, DummyJob, DummyProductJob, IncrementCounterJob


class FactoryTestCase(make_test_case_for_all_queues()):
    def setUp(self):
        super().setUp()
        DummyJob.objects.bulk_create([DummyJob(a=1, b=3) for _ in range(100)])
        DummyProductJob.objects.bulk_create(
            [
                DummyProductJob(
                    shop_name=shop_name,
                    product_name=product_name,
                    action=action,
                )
                for shop_name in ["shop-a", "shop-b", "shop-c", "shop-d"]
                for product_name in ["prod-a", "prod-b", "prod-c", "prod-d"]
                for action in ["FOO", "BAR", "BAZ"]
            ]
        )
        Counter.objects.create(counter_id="foo", value=0)
        IncrementCounterJob.objects.bulk_create(
            [IncrementCounterJob(counter_id="foo") for _ in range(100)]
        )


class MakeTestCaseForAllQueuesTests(TestCase):
    def test_covers_all_models_within_current_django_installation(self):
        self.assertEqual(
            ["DummyJob", "IncrementCounterJob", "DummyProductJob"],
            FactoryTestCase.queues,
        )

    def test_appends_static_fields_properly(self):
        WithoutStaticFields = make_test_case_for_all_queues()
        self.assertFalse(hasattr(WithoutStaticFields, "some_arbitrary_field"))

        WithStaticFields = make_test_case_for_all_queues(
            some_arbitrary_field=1337,
        )
        self.assertEqual(1337, WithStaticFields.some_arbitrary_field)



from unittest.mock import patch

from django.test import TestCase

from parviraptor.monitoring import QueueMonitoringResult, monitor_queue_entries

from .models import DummyJob, DummyProductJob


class MonitoringTests(TestCase):
    def test_does_not_complain_if_queue_is_empty(self):
        result = monitor_queue_entries([DummyJob])
        self.assertEqual(0, len(result))

    def test_complains_about_failed_jobs(self):
        with patch(
            "tests.models.DummyJob.count_failed_jobs",
            lambda: 1,
        ):
            result = monitor_queue_entries([DummyJob])
        self.assertEqual(1, len(result))

        self.assertEqual("DummyJob", result[0].queue_name)
        self.assertEqual(1, result[0].failed_jobs_count)
        self.assertEqual(0, result[0].long_processing_jobs_count)
        self.assertEqual(0, result[0].long_unprocessed_jobs_count)

    def test_complains_about_long_processing_jobs(self):
        with patch(
            "tests.models.DummyJob.count_long_processing_jobs",
            lambda: 1,
        ):
            result = monitor_queue_entries([DummyJob])
        self.assertEqual(1, len(result))

        self.assertEqual("DummyJob", result[0].queue_name)
        self.assertEqual(0, result[0].failed_jobs_count)
        self.assertEqual(1, result[0].long_processing_jobs_count)
        self.assertEqual(0, result[0].long_unprocessed_jobs_count)

    def test_complains_about_long_unprocessed_jobs(self):
        with patch(
            "tests.models.DummyJob.count_long_unprocessed_jobs",
            lambda: 1,
        ):
            result = monitor_queue_entries([DummyJob])
        self.assertEqual(1, len(result))

        self.assertEqual("DummyJob", result[0].queue_name)
        self.assertEqual(0, result[0].failed_jobs_count)
        self.assertEqual(0, result[0].long_processing_jobs_count)
        self.assertEqual(1, result[0].long_unprocessed_jobs_count)

    def test_flexible_long_unprocessed_timeframe(self):
        DummyJob.objects.create(a=1, b=2)
        self.assertEqual([], monitor_queue_entries([DummyJob]))
        with patch(
            "tests.models.DummyJob.MAX_TIMEFRAME_FOR_UNPROCESSED_JOBS_IN_MIN",
            -1,
        ):
            self.assertEqual(
                1, self._first_monitoring_entry.long_unprocessed_jobs_count
            )

    def test_flexible_long_processing_timeframe(self):
        DummyJob.objects.create(a=1, b=2, status="PROCESSING")
        self.assertEqual([], monitor_queue_entries([DummyJob]))
        with patch(
            "tests.models.DummyJob.MAX_TIMEFRAME_FOR_JOB_PROCESSING_IN_MIN", -1
        ):
            self.assertEqual(
                1, self._first_monitoring_entry.long_processing_jobs_count
            )

    def test_aggregates_results_across_multiple_jobs(self):
        DummyJob.objects.create(a=1, b=2, status="PROCESSING")
        with patch(
            "tests.models.DummyJob.MAX_TIMEFRAME_FOR_JOB_PROCESSING_IN_MIN",
            -1,
        ):
            with patch(
                "tests.models.DummyProductJob.count_long_unprocessed_jobs",
                lambda: 1337,
            ):
                result = monitor_queue_entries([DummyJob, DummyProductJob])
                self.assertEqual(
                    [
                        QueueMonitoringResult(
                            queue_name="DummyJob",
                            failed_jobs_count=0,
                            long_processing_jobs_count=1,
                            long_unprocessed_jobs_count=0,
                        ),
                        QueueMonitoringResult(
                            queue_name="DummyProductJob",
                            failed_jobs_count=0,
                            long_processing_jobs_count=0,
                            long_unprocessed_jobs_count=1337,
                        ),
                    ],
                    result,
                )

    @property
    def _first_monitoring_entry(self):
        result = monitor_queue_entries([DummyJob])
        self.assertEqual(1, len(result))
        return result[0]



from itertools import chain, islice
from typing import List

from django.apps import apps


def enumerate_job_models() -> List[type]:
    """
    Enumerates all non-abstract parviraptor job models within current
    Django environment.
    """
    # avoid circular imports
    from parviraptor.models.abstract import AbstractJob

    all_apps = apps.get_app_configs()
    relevant_models = list(chain(*(app.get_models() for app in all_apps)))

    def is_abstract(model_class):
        if not hasattr(model_class, "Meta"):
            # If there is no meta class, `abstract = True` can not have
            # been explicitly set so the model is logically non-abstract
            return False
        return not getattr(model_class.Meta, "abstract", False)

    non_abstract_models = list(filter(is_abstract, relevant_models))
    job_models = list(
        filter(
            lambda model_class: issubclass(model_class, AbstractJob),
            non_abstract_models,
        )
    )

    return job_models


def iter_chunks(size, iterable):
    """Divides `iterable` into tuples of `size`. The last chunk may be shorter.

    >>> list(iter_chunks(3, range(14)))
    [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9, 10, 11), (12, 13)]

    source: https://stackoverflow.com/a/22045226
    """
    it = iter(iterable)
    return iter(lambda: tuple(islice(it, size)), ())



from django.urls import path

from .views import OpenQueueEntriesView, queue_monitoring

urlpatterns = [
    path("queue-monitoring/", queue_monitoring, name="queue-monitoring"),
    path(
        "open-queue-entries/",
        OpenQueueEntriesView.as_view(),
        name="open-queue-entries",
    ),
]



import os

HELPER_SETTINGS = {
    "INSTALLED_APPS": ["parviraptor", "tests"],
    "ROOT_URLCONF": "parviraptor.urls",
    "DATABASES": {
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "HOST": os.environ["DB_HOST"],
            "NAME": os.environ["DB_NAME"],
            "PASSWORD": os.environ["DB_PASSWORD"],
            "USER": os.environ["DB_USER"],
            "TEST": {
                "CHARSET": "utf8",
                "COLLATION": "utf8_general_ci",
            },
        }
    },
}



from parviraptor.test import QueueTestCase

from .models import Counter, DummyJob, DummyProductJob, IncrementCounterJob


class ParallelityTests(QueueTestCase):
    """Schließt Nebenläufigkeitsprobleme bei parallelen Job-Queues aus."""

    def test_status_transition_from_new_to_processing_is_atomic(self):
        COUNTER_VALUE = 500
        jobs = [
            IncrementCounterJob(counter_id="foo") for _ in range(COUNTER_VALUE)
        ]

        # Ohne weitere Sperrmechanismen auf der Datenbank wird es dazu
        # kommen, dass voneinander unabhängige Jobs doppelt abgearbeitet
        # werden können. Wir arbeiten die Queue parallel ab und stellen sicher,
        # dass der Counter auf exakt n erhöht wurde.
        # Wäre er höher, so bedeutet das, dass Jobs doppelt abgearbeitet wurden.
        Counter.objects.create(counter_id="foo", value=0)
        self.process_queue(IncrementCounterJob, jobs, 8)
        self.assertEqual(
            COUNTER_VALUE, Counter.objects.get(counter_id="foo").value
        )

    def test_processes_strict_fifo_queue_in_right_order(self):
        jobs = [DummyJob(a=1, b=3) for _ in range(100)]
        self.process_queue(DummyJob, jobs, 8)

        ordered_ids = self.get_ordered_ids(DummyJob.objects.all(), "pk")
        ids_in_order_of_processing = self.get_ordered_ids(
            DummyJob.objects.all(), "modification_date"
        )
        self.assertEqual(ordered_ids, ids_in_order_of_processing)

    def test_processes_jobs_with_field_based_dependencies_in_right_order(self):
        shops = ["shop-a", "shop-b", "shop-c", "shop-d"]
        products = ["prod-a", "prod-b", "prod-c", "prod-d"]

        jobs = [
            DummyProductJob(
                shop_name=shop_name,
                product_name=product_name,
                action=action,
            )
            for shop_name in shops
            for product_name in products
            for action in range(1, 11)
        ]
        self.process_queue(DummyProductJob, jobs, 8)

        for shop_name in shops:
            for product_name in products:
                jobs = DummyProductJob.objects.filter(
                    shop_name=shop_name, product_name=product_name
                )
                ordered_ids = self.get_ordered_ids(jobs, "pk")
                ids_in_order_of_processing = self.get_ordered_ids(
                    jobs, "modification_date"
                )
                self.assertEqual(ordered_ids, ids_in_order_of_processing)



import logging
import signal
import threading
import traceback
from datetime import timedelta

from .exceptions import (
    DeferJob,
    IgnoreJob,
    InvalidJobError,
    TemporaryJobFailure,
    UnprocessableJob,
)
from .models.abstract import AbstractJob, BackoffStrategy, JobStatus

logger = logging.getLogger(__name__)

DEFAULT_TEMPORARY_FAILURE_THRESHOLD = 19


class QueueWorkerLogger:
    def __init__(self):
        self._counter = 0
        self._is_idle = False
        self._is_processable = True

    def info(self, *args, **kwargs):
        logger.info(*args, **kwargs)

    def debug(self, *args, **kwargs):
        logger.debug(*args, **kwargs)

    def mutate_to_processing_state(self):
        self._is_idle = False
        self._is_processable = True
        if self._counter == 0:
            self.info("processing jobs...")
        self._counter += 1

    def mutate_to_idle_state(self):
        if self._counter > 0 or not self._is_idle:
            self.info(f"processed {self._counter} jobs. awaiting new jobs.")
        self._is_idle = True
        self._counter = 0

    def mutate_to_unprocessable_state(self):
        if self._counter > 0 or self._is_processable:
            self.info(f"processed {self._counter} jobs so far.")
            self.info(
                "queue turned unprocessable right now. "
                "waiting until queue is processable again."
            )
        self._is_processable = False
        self._counter = 0


class QueueWorker:
    """Processing a certain job queue.

    Example:

    >>> worker = QueueWorker(DummyJob)
    >>> worker.run()  # blocks infinitely
    """

    def run(self):
        while not self._caught_exit_signal.is_set():
            try:
                job_worker = self._get_next_job_and_update_status()
                self.logger.mutate_to_processing_state()
                job_worker.process()
            except self.Job.DoesNotExist:
                self.logger.mutate_to_idle_state()
                self._sleep(self.pause_if_queue_empty)
            except UnprocessableJob:
                self.logger.mutate_to_unprocessable_state()
                self._sleep(self.pause_if_queue_empty)
            except TemporaryJobFailure as e:
                # this would not be reraised if `error_count` was reached.
                # see JobWorker.
                minutes = self._calc_latency_in_minutes(e.error_count)
                self._sleep(timedelta(minutes=minutes))

    def _calc_latency_in_minutes(self, error_count: int) -> int:
        match self.Job.BACKOFF_STRATEGY:
            case BackoffStrategy.CONSTANT:
                return self.Job.BACKOFF_STEP_MINUTES
            case BackoffStrategy.EXPONENTIAL:
                return self.Job.BACKOFF_STEP_MINUTES ** min(error_count, 5)

    def _get_next_job_and_update_status(self):
        job = self.Job.fetch_next_job()
        if job is None:
            raise self.Job.DoesNotExist()
        elif not job.is_processable():
            job.status = JobStatus.NEW
            job.save()
            raise UnprocessableJob()
        else:
            return JobWorker(
                job,
                temporary_failure_threshold=self.temporary_failure_threshold,
            )

    def __init__(
        self,
        Job: type[AbstractJob],
        pause_if_queue_empty=timedelta(minutes=1),
        temporary_failure_threshold=DEFAULT_TEMPORARY_FAILURE_THRESHOLD,
    ):
        self.Job = Job
        self.pause_if_queue_empty = pause_if_queue_empty
        self.temporary_failure_threshold = temporary_failure_threshold
        self.logger = QueueWorkerLogger()
        self._setup_signal_handling()

    def _setup_signal_handling(self):
        self._caught_exit_signal = threading.Event()
        for sig in [signal.SIGINT, signal.SIGTERM]:
            signal.signal(sig, self._exit_gracefully)

    def _exit_gracefully(self, signum, frame):
        self.logger.info(f"Exit signal caught. ({signal.Signals(signum).name})")
        self._caught_exit_signal.set()

    def _sleep(self, duration: timedelta):
        seconds = duration.total_seconds()
        self.logger.debug(f"... sleeping for {seconds}s")
        self._caught_exit_signal.wait(seconds)


class JobWorker:
    def __init__(
        self,
        job,
        temporary_failure_threshold=DEFAULT_TEMPORARY_FAILURE_THRESHOLD,
    ):
        if job.status != JobStatus.PROCESSING:
            raise ValueError(f"expected job to be PROCESSING, got {job.status}")
        self.job = job
        self.temporary_failure_threshold = temporary_failure_threshold

    def process(self):
        try:
            self.job.process()
            self.job.status = JobStatus.PROCESSED
        except DeferJob as e:
            self._info(f"Deferring job: {e}")
            self.job.status = JobStatus.DEFERRED
            self.job.error_message = str(e)
        except IgnoreJob as e:
            self._info(f"Ignoring job: {e}")
            self.job.status = JobStatus.IGNORED
            self.job.error_message = str(e)
        except InvalidJobError as e:
            self._info(f"Invalid job: {e}")
            self.job.status = JobStatus.FAILED
            self.job.error_message = str(e)
            self._log_status()
        except TemporaryJobFailure as e:
            self._warn(f"temporary failure: {e}")
            self.job.error_count += 1
            if self.job.error_count > self.temporary_failure_threshold:
                msg = "error count reached threshold"
                self._error(msg)
                self.job.status = JobStatus.FAILED
                self.job.error_message = msg
                self._log_status()
            else:
                self.job.status = JobStatus.NEW
                # we intentionally reraise this failure because the retry
                # backoff handling is done beyond the scope of this context
                raise
        except Exception as e:
            logger.error(self._format_log_message(str(e)))
            logger.error(traceback.format_exc())
            self.job.status = JobStatus.FAILED
            self.job.error_message = str(e)
            self._log_status()
        finally:
            self.job.save()

    def _log_status(self):
        self._info(f"status={self.job.status}")

    def _info(self, message: str):
        logger.info(self._format_log_message(message))

    def _warn(self, message: str):
        logger.warning(self._format_log_message(message))

    def _error(self, message: str):
        logger.error(self._format_log_message(message))

    def _format_log_message(self, message: str):
        return f"{type(self.job).__name__} {self.job.id}: {message}"



import traceback
from threading import Thread
from typing import Iterable

from django.test import TransactionTestCase

from ..models.abstract import AbstractJob
from .worker import InfinityLoopFreeQueueWorker


class QueueTestCase(TransactionTestCase):
    def process_queue(
        self,
        JobClass: type,
        jobs: Iterable[AbstractJob],
        number_of_threads: int,
        create_jobs: bool = True,
    ):
        """Processes a queue, optionally in parallel.

        If `number_of_threads >= 1`, they are parallelized using threads.

        This method creates the given job instances and ensures that in the end
        they are successfully processed.
        It does NOT block infinitely as the standard queue processor would do.
        """
        jobs = list(jobs)
        if create_jobs:
            JobClass.objects.bulk_create(jobs)
        self.assertEqual(
            (len(jobs), 0, 0, 0),
            (
                JobClass.objects.filter(status="NEW").count(),
                JobClass.objects.filter(status="PROCESSING").count(),
                JobClass.objects.filter(status="PROCESSED").count(),
                JobClass.objects.filter(status="FAILED").count(),
            ),
        )

        if number_of_threads == 1:
            InfinityLoopFreeQueueWorker(JobClass).run()
        else:
            self._process_concurrently(JobClass, number_of_threads)

        self.assertEqual(
            (0, 0, len(jobs), 0),
            (
                JobClass.objects.filter(status="NEW").count(),
                JobClass.objects.filter(status="PROCESSING").count(),
                JobClass.objects.filter(status="PROCESSED").count(),
                JobClass.objects.filter(status="FAILED").count(),
            ),
        )

    def _process_concurrently(self, JobClass, number_of_threads):
        class QueueWorkerThread(Thread):
            def __init__(self):
                super().__init__()
                self.succeeded = False
                self.failure_detail = None

            def run(self):
                try:
                    InfinityLoopFreeQueueWorker(JobClass).run()
                    self.succeeded = True
                except Exception as ex:
                    print(traceback.format_exc())
                    self.failure_detail = f"{type(ex).__name__}: {ex}"

        threads = [QueueWorkerThread() for _ in range(number_of_threads)]
        # join() blocks the current thread, this is why we need separate loops
        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()
        for idx, thread in enumerate(threads):
            if not thread.succeeded:
                if thread.failure_detail:
                    self.fail(
                        f"thread #{idx} failed with {thread.failure_detail}"
                    )
                else:
                    self.fail(f"thread #{idx} was however not processed")

    def get_ordered_ids(self, qs, field):
        return [job.pk for job in qs.order_by(field)]



from contextlib import contextmanager
from cProfile import Profile
from pstats import SortKey, Stats
from sys import stdout


@contextmanager
def measure_stats():
    pr = Profile()
    pr.enable()
    yield
    pr.disable()
    Stats(pr, stream=stdout).sort_stats(SortKey.CUMULATIVE).print_stats()



import enum
import itertools
from datetime import datetime, timedelta, timezone
from functools import reduce
from typing import Any

from django.db import models
from django.db.models import Q

from parviraptor.exceptions import TemporaryJobFailure


class JobStatus(models.TextChoices):
    NEW = "NEW"
    PROCESSING = "PROCESSING"
    PROCESSED = "PROCESSED"
    SQUASHED = "SQUASHED"
    FAILED = "FAILED"
    IGNORED = "IGNORED"
    DEFERRED = "DEFERRED"


@enum.unique
class BackoffStrategy(enum.Enum):
    EXPONENTIAL = enum.auto()
    CONSTANT = enum.auto()


class AbstractJob(models.Model):
    """Basisklasse zum Bilden einer Job-Queue.

    One should not directly derive from this class but use the base
    class inferred by `AbstractJobFactory`.

    Overridable class members:
    - MAX_TIMEFRAME_FOR_JOB_PROCESSING_IN_MIN
    - MAX_TIMEFRAME_FOR_JOB_UNPROCESSED_IN_MIN
    - BACKOFF_STEP_MINUTES
    - BACKOFF_STRATEGY
    """

    MAX_TIMEFRAME_FOR_JOB_PROCESSING_IN_MIN = 30
    MAX_TIMEFRAME_FOR_UNPROCESSED_JOBS_IN_MIN = 16 * 60

    BACKOFF_STEP_MINUTES = 2
    BACKOFF_STRATEGY = BackoffStrategy.EXPONENTIAL

    @classmethod
    def get_dependent_fields(cls):
        if not hasattr(cls, "dependent_fields"):
            raise AttributeError(
                f"{cls} is somehow misconfigured: "
                + "missing 'dependent_fields'"
            )
        return getattr(cls, "dependent_fields")

    Status = JobStatus

    creation_date = models.DateTimeField(
        auto_now_add=True,
    )
    modification_date = models.DateTimeField(
        auto_now=True,
    )
    status = models.CharField(
        choices=Status.choices,
        db_index=True,
        default=Status.NEW,
        max_length=32,
    )
    error_count = models.IntegerField(
        default=0,
    )
    error_message = models.TextField(
        blank=True,
        null=True,
    )

    def process(self):
        """Processes a job.

        Derived classes implement their concrete job logic here. This
        implementation does not have to care about concrete state
        transitions.

        One can raise any exception which cause a job to finalize to the
        `FAILED` state. Raising any `TemporaryJobFailure` triggers the
        job to be retried if it has not failed too often.

        In case job processing fails you might need to implement reasonable
        rollback mechanisms. For mutating database state, you might e.g.
        want to use `transaction.atomic` as decorator / context manager.
        """
        raise NotImplementedError()

    def is_processable(self) -> bool:
        """Returns whether the job queue is processable.

        By default, it always returns True. This method may be overriden
        in case the processability of a certain job queue depends on
        runtime condition. If 'False' is returned, the job worker processing
        this job class will silently await until 'True' is returned again.
        """
        return True

    @classmethod
    def fetch_next_job(cls, id_gt=None):
        job = cls._fetch_next_new_job(id_gt=id_gt)
        if job is None:
            raise cls.DoesNotExist()

        if cls.get_dependent_fields() is not None:
            # If current job depends on failed predecessors, we can set this
            # job and all of its successors to FAILED as well.
            failed_predecessors = cls._fetch_failed_predecessors(job)
            cls._change_job_and_dependent_successors_to_failed_if_necessary(
                job, failed_predecessors
            )

            # We try to process the next direct successor on incomplete or
            # failed predecessors as the next one might be processable.
            if (
                cls._incomplete_predecessors_exist(job)
                or failed_predecessors.exists()
            ):
                return cls.fetch_next_job(id_gt=job.id)

        # concurrency mitigation: two parallel workers might set the same job
        # to PROCESSING at the same time. `updated_jobs_count == 0` means
        # "another process already took care of this job". In this case we
        # can just silently continue.
        updated_jobs_count = cls.objects.filter(
            status=JobStatus.NEW, id=job.id
        ).update(
            status=JobStatus.PROCESSING,
            modification_date=datetime.now(tz=timezone.utc),
        )

        if updated_jobs_count == 1:
            job.status = JobStatus.PROCESSING
            return job
        elif updated_jobs_count == 0:
            return cls.fetch_next_job()
        else:
            raise RuntimeError()

    @classmethod
    def _fetch_next_new_job(cls, id_gt=None):
        jobs = cls.objects.filter(status=JobStatus.NEW)
        if id_gt is not None:
            jobs = jobs.filter(id__gt=id_gt)
        return jobs.order_by("id").first()

    @classmethod
    def _get_dependent_fields_lookup(cls, job):
        return reduce(
            lambda combined, field: combined
            & Q(**{field: getattr(job, field)}),
            cls.get_dependent_fields(),
            Q(),
        )

    @classmethod
    def _incomplete_predecessors_exist(cls, job):
        return cls.objects.filter(
            cls._get_dependent_fields_lookup(job),
            status__in=("NEW", "PROCESSING"),
            id__lt=job.id,
        ).exists()

    @classmethod
    def _fetch_failed_predecessors(cls, job):
        return cls.objects.filter(
            cls._get_dependent_fields_lookup(job),
            status="FAILED",
            id__lt=job.id,
        )

    @classmethod
    def _change_job_and_dependent_successors_to_failed_if_necessary(
        cls, job, failed_predecessors
    ):
        if failed_predecessors.exists():
            cls.objects.filter(
                cls._get_dependent_fields_lookup(job),
                id__gte=job.id,
            ).update(
                status="FAILED",
                error_message="dependent jobs failed",
            )

    @classmethod
    def get_queryset_filters_for_disjoint_queues(
        cls,
    ) -> list[dict[str, Any]]:
        dependent_fields = cls.get_dependent_fields()
        if dependent_fields is None:
            return []
        combinations = itertools.product(
            *[
                set(cls.objects.values_list(field, flat=True))
                for field in dependent_fields
            ]
        )
        filters = [
            dict(zip(dependent_fields, combination))
            for combination in combinations
        ]
        return filters

    @classmethod
    def count_failed_jobs(cls) -> int:
        return cls.objects.filter(status=cls.Status.FAILED).count()

    @classmethod
    def count_long_processing_jobs(cls) -> int:
        dt = datetime.now(tz=timezone.utc) - timedelta(
            minutes=cls.MAX_TIMEFRAME_FOR_JOB_PROCESSING_IN_MIN
        )

        return cls.objects.filter(
            status=cls.Status.PROCESSING,
            modification_date__lt=dt,
        ).count()

    @classmethod
    def count_long_unprocessed_jobs(cls) -> int:
        dt = datetime.now(tz=timezone.utc) - timedelta(
            minutes=cls.MAX_TIMEFRAME_FOR_UNPROCESSED_JOBS_IN_MIN
        )

        return cls.objects.filter(
            status=cls.Status.NEW,
            creation_date__lt=dt,
        ).count()

    def raise_temporary_failure(self, message: str):
        """Raises a temporary failure.

        Temporary failures are such which might not occur after one or more
        retries. `process()` should call this method so the job can be resumed
        by the worker in case the temporary failure threshold has not been
        exceeded.
        """
        raise TemporaryJobFailure(message, self.error_count)

    class Meta:
        abstract = True


class AbstractJobFactory:
    """Factory for inferring job queue base classes.

    `dependent_fields` is a list of model field names. If there are multiple
    jobs with the same field values, they are processed by FIFO strategy. This
    enables for disjoint queues within one queue model.

    - `dependent_fields = []` declares a strict FIFO queue.
    - `dependent_fields = None` means all jobs can be processed from each
      other independently.

    "FIFO strategy" also covers that successors with same `dependent_fields`
    values are only processed if all pending predecessors have been
    *successfully* completed.
    """

    @classmethod
    def make_base_class(
        cls, dependent_fields: list[str] | None
    ) -> type[AbstractJob]:
        class DerivedJob(AbstractJob):
            def __init_subclass__(cls):
                # `cls` is a derived class
                setattr(cls, "dependent_fields", dependent_fields)

            class Meta(AbstractJob.Meta):
                abstract = True

        return DerivedJob



from setuptools import find_packages, setup

from parviraptor import __version__

setup(
    name="parviraptor",
    version=__version__,
    description="Django-based job queue",
    author="puzzleYOU GmbH",
    packages=find_packages("."),
    install_requires=[
        "Django<5",
    ],
    zip_safe=True,
)



class DeferJob(Exception):
    pass


class IgnoreJob(Exception):
    pass


class UnprocessableJob(Exception):
    pass


class InvalidJobError(Exception):
    pass


class TemporaryJobFailure(Exception):
    def __init__(self, message: str, error_count: int):
        self._message = message
        self._error_count = error_count

    def __str__(self):
        return self._message

    @property
    def message(self):
        return self._message

    @property
    def error_count(self):
        return self._error_count



from django.shortcuts import render

from ..monitoring import monitor_queue_entries
from ..utils import enumerate_job_models


def queue_monitoring(request):
    job_classes = enumerate_job_models()
    queue_entries = monitor_queue_entries(job_classes)
    context = {"queue_entries": queue_entries}
    return render(request, "parviraptor/queue_monitoring.html", context)



from unittest.mock import Mock

from ..worker import QueueWorker


class InfinityLoopFreeQueueWorker(QueueWorker):
    """
    A non-blocking `QueueWorker` implementation, primarily for automated tests.

    `QueueWorker.run()` by default blocks until there are new processable jobs
    or an exit signal has been caught. This implementation terminates after
    all currently processable jobs have been completed.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._should_stop = False

    @property
    def _caught_exit_signal(self):
        return Mock(is_set=lambda: self._should_stop)

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

    def _sleep(self, _):
        self._should_stop = True

    def _setup_signal_handling(self):
        pass



import signal
from random import randint
from time import sleep

from django.db import models, transaction

from parviraptor.exceptions import DeferJob, IgnoreJob, InvalidJobError
from parviraptor.models import AbstractJobFactory

MAX_ERROR_COUNT = 5


class DummyJob(AbstractJobFactory.make_base_class([])):
    """Beispiel-Job zu Demonstrations- und Testzwecken."""

    a = models.IntegerField()
    b = models.IntegerField()
    result = models.IntegerField(
        null=True,
    )

    def process(self):
        sleep_randomly()

        if self.a == 0 and self.b == 100000:
            self.raise_temporary_failure("always fail temporarily")

        self.result = self.a + self.b

        # Normalerweise kommen die Signals von außerhalb. Zum Testen ist es
        # aber leichter, wenn wir sie deterministisch selbst senden können, und
        # zwar während der Verarbeitung eines Jobs (d.h. "in" `Job.process()`,
        # also hier:
        if self.result < 0:
            signal.raise_signal(signal.SIGTERM)

        # Willkürliches Beispiel für TemporaryJobFailure, i.d.R. sollte
        # `process` nur auf die hier definierten Felder zugreifen müssen, und
        # den `error_count` komplett ignorieren.
        if self.a == 0 and self.error_count < MAX_ERROR_COUNT:
            self.raise_temporary_failure("adding to 0 failed")
        elif self.b == 0:
            raise ValueError("b cannot be 0")
        elif self.result == 100:
            raise InvalidJobError(f"Ignoring result {self.result}")
        elif self.result == 200:
            raise IgnoreJob(f"Ignoring result {self.result}")
        elif self.result == 300:
            raise DeferJob(f"Deferring result {self.result}")


class Counter(models.Model):
    counter_id = models.CharField(max_length=16, primary_key=True)
    value = models.IntegerField()


class IncrementCounterJob(AbstractJobFactory.make_base_class(None)):
    counter_id = models.CharField(max_length=16)

    @transaction.atomic
    def process(self):
        state = (
            Counter.objects.select_for_update()
            .filter(counter_id=self.counter_id)
            .first()
        )
        state.value += 1
        state.save()


class DummyProductJob(
    AbstractJobFactory.make_base_class(["shop_name", "product_name"])
):
    shop_name = models.CharField(max_length=16, db_index=True)
    product_name = models.CharField(max_length=16, db_index=True)
    action = models.CharField(max_length=16)

    def process(self):
        sleep_randomly()


def sleep_randomly():
    sleep(0.001 * randint(0, 10))
